unit files;
{ Open and close files, interpret command line, do input}

interface  uses globals;

procedure readParagraph (var P: paragraph; var no: line_nos;
  var L: paragraph_index0);
function styleFileFound: boolean;
procedure processOption(j: char);
procedure showOptions(anyway: boolean);
procedure OpenFiles;
procedure CloseFiles;
procedure putLine (line: string);
procedure put(line: string; putspace: boolean);
procedure tex3 (s: string);
function endOfInfile: boolean;

implementation uses strings;

const  param_leader = '-';
       stylefilename: string = 'mtxstyle.txt';

var buffer: string;
    infilename, outfilename: string;
    infilecount, teststyle: integer;
    eof_infile: boolean;

function endOfInfile: boolean;
  begin endOfInfile:=eof_infile; end;

procedure tex3 (s: string);
  begin
    if pmx_preamble_done then
      if s[1]='%' then putLine(s)
      else if first_paragraph then putLine('\'+s+'\')
      else putLine('\\'+s+'\')
    else putLine(s);
  end;

procedure putLine (line: string);
  begin  if outlen+length(line) > pmxlinelength-1 then writeln(outfile);
    writeln(outfile,line); outlen:=0;
  end;

procedure put(line: string; putspace: boolean);
  var l: integer;
  begin l:=length(line);
    if l>pmxlinelength then error('Item longer than PMX line length',print);
    if outlen+l > pmxlinelength-1 then
    begin putLine(''); put(line,false) end
    else begin if putspace and (outlen>0) and (line[1]<>' ') then
        line:=' '+line;
      write(outfile,line); inc(outlen,l);
    end
  end;

procedure showOptions(anyway: boolean);
  begin
     if anyway then writeln ('  Options available:')
       else writeln ('  Options selected:');
     if anyway or unbeam_if_vocal then 
        writeln('Never implicitly unbeam short notes (b)');
     if anyway or not dochords then writeln ('Ignore chord lines (c)');
     if anyway or tonic_solfa then writeln('Note names in tonic sol-fa (f)');
     if anyway or ignore then writeln ('Ignore errors (i)');
     if anyway or not dolyrics then writeln ('Music only, no lyrics (m)');
     if anyway or set_names then writeln('Set instrument names (n)');
     if not anyway and newslurs then writeln ('Use () slurs in PMX file (s)');
     if anyway or not dotext then writeln ('Ignore uptext lines (t)');
     if anyway or no_uptext_for_rests then writeln ('Uptext on notes only (u)');
     if anyway or (verbose=1) then writeln ('Verbose progress report (v)');
     if anyway or (verbose=2) then writeln ('Debugging printouts (D)');
     if anyway or pedantic then writeln ('Suppress pedantic warnings (w)');
     if anyway or extended_dots then 
       writeln('Enable extended dot positioning syntax (x)');
     if anyway then writeln('Chosen case = [0123456789]')
       else if (choice<>' ') then writeln('Chosen case = ',choice);
  end;

function styleFileFound: boolean;
begin styleFileFound := teststyle<>0; end;

procedure processOption(j: char);
  begin  case j of
  'b': unbeam_if_vocal:=false;
  'c': dochords:=false;
  'd': warning('-d option ignored, not supported past M-Tx 0.30',not print);
  'f': tonic_solfa:=true;
  'i': ignore:=true;
  'l': warning('-l option ignored, not supported past M-Tx 0.30',not print);
  'm': dolyrics:=false;
  'n': set_names:=true;
  'p': private_pmx:=true;
  's': newslurs:=true;
  't': dotext:=false;
  'u': no_uptext_for_rests:=true;
  'v': verbose:=1;
  'w': pedantic:=false;
  'x': extended_dots:=true;
  'D': verbose:=2;
  '0'..'9': choice:=j;
  else showOptions(true);
    write(j); error(': invalid option',not print);
  end;
end;

procedure OpenFiles;
var i, j, l, fileid, testin: integer;
    basename, param, ext: string;
  procedure checkExistingFile;
    var tryfile: file;
  begin
{$I-}
    assign(tryfile,basename);  reset(tryfile);  testin := ioresult;
{$I+}
    if testin<>0 then exit else close(tryfile);
    writeln('There exists a file named ',basename,'.  I am treating this');
    error('  as a fatal error unless you specify -i',not print);
  end;
begin
  fileid:=0;  infilecount:=0;  line_no:=0;  paragraph_no:=0;
  for i:=1 to ParamCount do
  begin param:=ParamStr(i);
    if param[1]=param_leader then
      for j:=2 to length(param) do processOption(param[j])
    else if fileid=0 then fileid:=i
    else if texdir='' then texdir:=param
    else stylefilename:=param;
  end;
  if fileid=0 then fatalError('You must supply a basename on the command line')
  else basename:=paramstr(fileid);
  l:=length(basename);
  if (l>4) and (basename[l-3]='.') then
  begin ext:=substr(basename,l-2,3); toUpper(ext); if ext='MTX' then
    begin warning('.mtx extension deleted from basename',not print);
      shorten(basename,l-4);
    end;
  end;
  if pos1('.',basename)>0 then checkExistingFile;
  infilename := basename+'.mtx';  outfilename := basename+'.pmx';
{$I-}
  assign(infile,infilename);  reset(infile);  testin := ioresult;
    buffer:='';
  assign(outfile,outfilename);  rewrite(outfile);
  assign(stylefile,stylefilename);  reset(stylefile);  teststyle := ioresult;
{$I+}
  if testin<>0 then fatalError('Input file '+infilename+' not found');
  outfile_open := true;  writeln('Writing to ',basename,'.pmx');
end;

procedure CloseFiles;
begin  close(outfile);  close(infile);
  if teststyle=0 then close(stylefile);
end;

function skipEmptyLines: boolean;  { TRUE if non-empty line found }
  var started: boolean;
begin
  if posnot(blank,buffer) = 0 then buffer := '';
  if buffer <> '' then begin skipEmptyLines:=true; exit; end;
  started := false;
  repeat
    eof_infile := eof(infile);
    if (buffer='') and not eof_infile then
    begin  readln (infile, buffer);  inc(infilecount);
      if posnot(blank,buffer) = 0 then buffer := ''
      else started := true;
    end
  until started or eof_infile;
  if eof_infile then buffer := '';
  skipEmptyLines := started;
end;

procedure readParagraph (var P: paragraph; var no: line_nos;
                         var L: paragraph_index0);
  var ended, dropline: boolean;
begin
  L:=0; if not skipEmptyLines then exit;
  inc(paragraph_no); dropline:=false;
  if startsWithIgnoreCase(buffer,'SUSPEND') then
    begin ignore_input:=true; dropline:=true; end;
  if startsWithIgnoreCase(buffer,'RESUME') then
    begin ignore_input:=false; dropline:=true; end;
  repeat
    if L<lines_in_paragraph then if dropline then dropline:=false else
      begin inc(L); P[L]:=buffer; buffer:=''; no[L]:=infilecount; end
    else begin line_no := infilecount;
       warning('Paragraph too long: skipping line',not print);
    end;
    eof_infile := eof(infile);
    if eof_infile then  buffer:=''
    else begin  readln(infile,buffer);  inc(infilecount);  end;
    ended:=(posnot(blank,buffer)=0) or eof_infile;
  until ended;
  ended := skipEmptyLines;
end;

end.
