unit status;
{ Keep track of duration, octave, slur and beam status. }

interface

procedure initStatus;
procedure saveStatus(voice: integer);

procedure resetDuration (voice: integer; dur: char);
function duration(voice: integer): char;

function slurLevel(voice: integer): integer;
function delaySlur(voice: integer): integer;
procedure renewSlurLevel(voice, dir: integer);
function afterBeam(voice: integer): integer;
function noBeamMelisma(voice: integer): boolean;
function noSlurMelisma(voice, history: integer): boolean;
procedure clearBeamSlurEnd(voice: integer);
function lastUnderBeam(voice: integer): integer;
procedure incLastUnderBeam(voice: integer);
function lastUnderSlur(voice: integer): integer;
procedure incLastUnderSlur(voice: integer);
function afterSlur(voice: integer): integer;
procedure setUnbeamed(voice: integer);
procedure setUnslurred(voice: integer);
procedure beginBeam(voice: integer; var note: string);
procedure beginSlur(voice: integer; var note: string);
procedure endSlur(voice: integer; var note: string);
procedure activateBeamsAndSlurs(voice: integer);
procedure appendSlur(voice: integer; var note: string);

procedure setOctave(voice: integer);
procedure resetOctave(voice: integer);
function octave(voice: integer): char;
procedure newOctave(voice: integer; dir: char);
procedure initOctaves(octaves: string);

procedure renewPitch (voice: integer; var note: string);
procedure renewChordPitch (voice: integer; note: string);
procedure rememberDurations;
procedure restoreDurations;

implementation uses globals, strings, mtxline;

const
  uppers = 'QWERTYUIOPASDFGHJKLZXCVBNM';
  init_oct: string = '';

type
  line_status = record
    pitch, chord_pitch, octave_adjust, after_beam, last_under_slur,
      last_under_beam, slur_level, after_slur: integer;
    octave, lastnote, chord_lastnote, duration: char;
    slur_toggle: array[1..2] of char;
    beamnext, beamed, slurnext, slurred, no_beam_melisma: boolean;
    no_slur_melisma: array[1..2] of boolean;
    delay_slur: string;
  end;

var current: array[voice_index] of line_status;
    lastdur: array[voice_index] of char;

procedure rememberDurations;
  var v: voice_index;
begin for v:=1 to nvoices do lastdur[v]:=duration(v) end;

procedure restoreDurations;
  var v: voice_index;
begin for v:=1 to nvoices do resetDuration(v,lastdur[v]) end;

function duration(voice: integer): char;
begin duration := current[voice].duration; end;

procedure resetDuration(voice: integer; dur: char);
begin if pos1(dur,durations)=0 then
    error3(voice,'M-Tx system error: resetDuration');
  current[voice].duration := dur
end;

procedure activateBeamsAndSlurs(voice: integer);
begin  with current[voice] do
  begin
    if beamnext then begin beamed:=true; beamnext:=false; end;
    if beamed then inc(after_beam);
    if slurnext then begin slurred:=true; slurnext:=false; end;
    if slurred then inc(after_slur);
  end
end;

procedure saveStatus(voice: integer);
begin  with current[voice] do
  begin chord_pitch := pitch;  chord_lastnote := lastnote; end;
end;

procedure appendSlur(voice: integer; var note: string);
begin  if not newslurs then
  with current[voice] do if delay_slur<>'' then
  begin
    if not print_all_slurs 
      and (endsWith(delay_slur,'s~') or endsWith(delay_slur,'t~')) then
      shorten(delay_slur,length(delay_slur)-2);
    if not print_all_slurs 
      and (startsWith(delay_slur,'s~') or startsWith(delay_slur,'t~')) then
      predelete(delay_slur,2);
    note := note + ' ' + delay_slur;
    delay_slur:='';
  end;
end;

function afterBeam(voice: integer): integer;
begin afterBeam := current[voice].after_beam; end;

function delaySlur(voice: integer): integer;
begin delaySlur:=wordCount(current[voice].delay_slur); end;

function slurLevel(voice: integer): integer;
begin slurLevel := current[voice].slur_level; end;

procedure renewSlurlevel(voice, dir: integer);
begin inc(current[voice].slur_level,dir); end;

function noBeamMelisma(voice: integer): boolean;
begin noBeamMelisma := current[voice].no_beam_melisma; end;

function noSlurMelisma(voice, history: integer): boolean;
begin with current[voice] do
  noSlurMelisma := no_slur_melisma[slur_level+history];
end;

procedure clearBeamSlurEnd(voice: integer);
begin with current[voice] do
  begin last_under_beam:=0; last_under_slur:=0; end;
end;

function lastUnderBeam(voice: integer): integer;
begin  lastUnderBeam:= current[voice].last_under_beam; end;

procedure incLastUnderBeam(voice: integer);
begin  inc(current[voice].last_under_beam); end;

function lastUnderSlur(voice: integer): integer;
begin lastUnderSlur := current[voice].last_under_slur; end;

function afterSlur(voice: integer): integer;
begin with current[voice] do afterSlur := after_slur; end;

procedure incLastUnderSlur(voice: integer);
begin  inc(current[voice].last_under_slur); end;

function octave(voice: integer): char;
begin octave := current[voice].octave; end;

procedure resetOctave(voice: integer);
begin current[voice].octave := ' '; end;

procedure initOctaves(octaves: string);
  var i: integer;
begin init_oct:=octaves;
  i:=1;
  while i<=length(init_oct) do
    if init_oct[i]=' ' then delete1(init_oct,i) else inc(i);
end;

function initOctave(voice_stave: stave_index): char;
begin  
  if voice_stave>length(init_oct) then  
  if pos1(clef[voice_stave],'Gt08')>0
  then initOctave:='4' else initOctave:='3'
  else initOctave:=init_oct[voice_stave];
end;

procedure setOctave(voice: integer);
begin current[voice].octave:=initOctave(voiceStave(voice)); end;

procedure newOctave(voice: integer; dir: char);
begin with current[voice] do case dir of
  '+': inc(octave);
  '-': dec(octave);
  end;
end;

function newPitch (note: string; pitch: integer; lastnote: char): integer;
  var i, l, interval, npitch: integer;
begin
  interval := ord(note[1])-ord(lastnote);
  if interval>3 then dec(interval,7);
  if interval<-3 then inc(interval,7);
  npitch:=pitch+interval;
  l:=length(note);
  for i:=1 to l do  if note[i]='+' then inc(npitch,7)
    else if note[i]='-' then dec(npitch,7);
  newPitch:=npitch;
end;

procedure repitch(var note: string; diff: integer);
  var l: integer;
  procedure delins(var note: string; c1, c2: char; l: integer);
    var i, n: integer;
  begin  n:=length(note); i:=pos1(c1,note); if i=0 then i:=n+1;
    while (l>0) and (i<=n) do begin delete1(note,i); dec(n); dec(l); end;
    i:=pos1(c2,note);
    if i=0 then if length(note)<2 then error('M-Tx program error',print)
      else i:=3;
    while l>0 do begin insertchar(c2,note,i); dec(l); end;
  end;
  begin  diff:=diff div 7;
    if diff>0 then delins(note, '-','+',diff)
    else delins(note, '+','-',-diff);
  end;

procedure setUnbeamed(voice: integer);
begin  with current[voice] do  begin beamed:=false; after_beam:=0; end;
end;

procedure setUnslurred(voice: integer);
begin with current[voice] do
  begin slurred:=false; after_slur:=0; end;
end;

procedure beginBeam(voice: integer; var note: string);
begin  with current[voice] do
  begin  if beamed then error3(voice,
    'Starting a forced beam while another is open');
    beamnext := true;  no_beam_melisma:=startsWith(note,'[[');
    if no_beam_melisma then predelete(note,1);
  end;
end;

{ The following arrays are not yet used.  Later we will test
  whether labelled slurs are opened and closed properly. }

var slur_open1: array['1'..'9'] of boolean;
    slur_openA: array['A'..'Z'] of boolean;

procedure clearSlurs;
  var c: char;
begin for c:='1' to '9' do slur_open1[c]:=false;
  for c:='A' to 'Z' do slur_openA[c]:=false;
end;

procedure beginSlur(voice: integer; var note: string);
  var actualslur: char;
      keepnote: string;
      labelled: boolean;
  procedure selectToggles;
  begin  with current[voice] do  case slur_level of
  1:  if note[1]=start_slur then slur_toggle[1]:='s'
            else slur_toggle[1]:='t';
  2:  if slur_toggle[1]='t' then slur_toggle[2]:='s'
            else slur_toggle[2]:='t';
    end;
  end;
begin  keepnote:=note;
  with current[voice] do
  begin
    if slur_level>1 then Error3(voice,'More than two levels of nested slurs');
    inc(slur_level);
    no_slur_melisma[slur_level] := startsWith(note,'((') or
      startsWith(note,'{{');
    if no_slur_melisma[slur_level] then predelete(note,1);
    if slur_level=2 then after_slur:=0;
    slurnext := true;
    labelled := false;
    if length(note)>1 then labelled := pos1(note[2],digits+uppers)>0;
    if labelled then slur_toggle[slur_level]:='s' else selectToggles;
    actualslur := note[1]; note[1]:=slur_toggle[slur_level];
{ Maybe delay_slur should not be used with labelled slurs }
    if delay_slur='' then delay_slur:=note
    else begin
      if (actualslur<>'{') and not labelled then
      begin slur_toggle[1]:='t'; slur_toggle[2]:='s'; end;
      note[1]:=slur_toggle[2]; delay_slur[1]:=slur_toggle[1];
      delay_slur:=delay_slur+' '+note;
    end;
    if note[1]='t' then
    if (pos1('+',note)>0) or (pos1('-',note)>0) then
            Error3(voice,'`t'' slurs may not have adjustments');
    if newslurs then note:=keepnote else note:='';
  end;
end;

procedure endSlur(voice: integer; var note: string);
  var s2: integer;
      keepnote: string;
begin  keepnote:=note;
  with current[voice] do
  begin
    if slur_level<1 then Error3(voice,'Ending a slur that was never started');
    note[1]:=slur_toggle[slur_level]; s2:=pos1(start_slur,note);
    if s2=0 then renewSlurLevel(voice,-1)
    else  begin note[s2]:=note[1]; insertchar(blank,note,s2);
    end;
    if slur_level=0 then setUnslurred(voice);
  end;
  if newslurs then note:=keepnote;
end;

procedure renewPitch (voice: integer; var note: string);
  var pstat: integer;
begin with current[voice] do
  begin  pstat:=newPitch (note, chord_pitch, chord_lastnote);
    pitch := newPitch (note, pitch, lastnote);
    if pitch<>pstat then repitch(note,pitch-pstat);
    lastnote:=note[1];
    if (verbose>2) and (pitch<>pstat) then
      writeln('Pitch from melodic line = ',pitch,
      '   from last chordal note = ', pstat);
  end;
end;

procedure renewChordPitch (voice: integer; note: string);
begin  with current[voice] do
  begin  chord_pitch:=newPitch(note,chord_pitch,chord_lastnote);
    chord_lastnote:=note[1];
  end
end;

procedure initStatus;
  var voice: integer;
begin  for voice:=1 to nvoices do with current[voice] do
  begin duration:=default_duration;
    lastnote:='f'; octave_adjust:=0; delay_slur:=''; slur_level:=0;
    beamed:=false;  beamnext:=false;  after_beam:=0;
    slurred:=false;  slurnext:=false;  after_slur:=0;
    clearSlurs;
    octave:=initOctave(voiceStave(voice));
    pitch:=7*(ord(octave)-ord('0'))-3;
    saveStatus(voice);
  end;
end;

end.
